from unittest import TestCase

import pandas as pd
from pandas.testing import assert_series_equal, assert_frame_equal

from datamatch.pairers import MatchPairer, DeduplicatePairer
from datamatch.indices import ColumnsIndex


class BasePairerTestCase(TestCase):
    def assert_pair_equal(self, pair_a, pair_b):
        self.assertEqual(len(pair_a), 2)
        self.assertEqual(len(pair_b), 2)
        for i, (name, ser) in enumerate(pair_a):
            self.assertEqual(name, pair_b[i][0])
            assert_series_equal(ser, pair_b[i][1], check_names=False)


class MatchPairerTestCase(BasePairerTestCase):
    def test_pairs(self):
        columns = ['x', 'y', 'z']
        dfa = pd.DataFrame([
            [1, 'a', 'b'],
            [2, 'c', 'd'],
            [3, 'e', 'f']
        ], columns=columns)
        dfb = pd.DataFrame([
            [1, 'q', 'w'],
            [4, 'z', 'x'],
            [2, 'a', 's'],
        ], columns=columns)

        pairer = MatchPairer(dfa, dfb, ColumnsIndex(['x']))
        assert_frame_equal(pairer.frame_a, dfa)
        assert_frame_equal(pairer.frame_b, dfb)

        pairs = list(pairer.pairs())
        self.assertEqual(len(pairs), 2)
        self.assert_pair_equal(pairs[0], (
            (0, pd.Series([1, 'a', 'b'], index=columns)),
            (0, pd.Series([1, 'q', 'w'], index=columns)),
        ))
        self.assert_pair_equal(pairs[1], (
            (1, pd.Series([2, 'c', 'd'], index=columns)),
            (2, pd.Series([2, 'a', 's'], index=columns)),
        ))


class DeduplicatePairerTestCase(BasePairerTestCase):
    def test_pairs(self):
        columns = ['x', 'y', 'z']
        df = pd.DataFrame([
            [1, 'a', 'b'],
            [2, 'c', 'd'],
            [3, 'e', 'f'],
            [1, 'q', 'w'],
            [4, 'z', 'x'],
            [2, 'a', 's'],
        ], columns=columns)

        pairer = DeduplicatePairer(df, ColumnsIndex(['x']))
        assert_frame_equal(pairer.frame_a, df)
        assert_frame_equal(pairer.frame_b, df)

        pairs = list(pairer.pairs())
        self.assertEqual(len(pairs), 2)
        self.assert_pair_equal(pairs[0], (
            (0, pd.Series([1, 'a', 'b'], index=columns)),
            (3, pd.Series([1, 'q', 'w'], index=columns)),
        ))
        self.assert_pair_equal(pairs[1], (
            (1, pd.Series([2, 'c', 'd'], index=columns)),
            (5, pd.Series([2, 'a', 's'], index=columns)),
        ))



import unittest
import pandas as pd
from pandas.core.indexes.range import RangeIndex
from pandas.testing import assert_frame_equal
import itertools

from datamatch.indices import MultiIndex, NoopIndex, ColumnsIndex


class BaseIndexTestCase(unittest.TestCase):
    def assert_pairs_equal(self, pair_a, pair_b):
        df1, df2 = pair_a
        df3, df4 = pair_b
        assert_frame_equal(df1, df3)
        assert_frame_equal(df2, df4)

    def assert_pairs_list_equal(self, list_a, list_b):
        self.assertEqual(len(list_a), len(list_b))
        for pair_a, pair_b in itertools.zip_longest(list_a, list_b):
            self.assert_pairs_equal(pair_a, pair_b)


class TestNoopIndex(BaseIndexTestCase):
    def test_index(self):
        df = pd.DataFrame([[1, 2], [3, 4]])
        idx = NoopIndex()
        keys = idx.keys(df)
        self.assertEqual(keys, set([0]))
        assert_frame_equal(idx.bucket(df, 0), df)


class TestColumnsIndex(BaseIndexTestCase):
    def test_index(self):
        cols = ["c", "d"]
        df = pd.DataFrame(
            [[1, 2], [2, 4], [3, 4]], index=["x", "y", "z"], columns=cols)
        idx = ColumnsIndex(["c"])
        keys = idx.keys(df)
        self.assertEqual(keys, set([(1,), (2,), (3,)]))
        assert_frame_equal(
            idx.bucket(df, (1,)),
            pd.DataFrame([[1, 2]], index=["x"], columns=cols)
        )
        assert_frame_equal(
            idx.bucket(df, (2,)),
            pd.DataFrame([[2, 4]], index=["y"], columns=cols)
        )
        assert_frame_equal(
            idx.bucket(df, (3,)),
            pd.DataFrame([[3, 4]], index=["z"], columns=cols)
        )

    def test_multi_columns(self):
        cols = ["c", "d"]
        df = pd.DataFrame(
            [[1, 2], [2, 4], [3, 4]], index=["z", "x", "c"], columns=cols)
        idx = ColumnsIndex(["c", "d"])
        keys = idx.keys(df)
        self.assertEqual(keys, set([(1, 2), (2, 4), (3, 4)]))
        assert_frame_equal(
            idx.bucket(df, (1, 2)),
            pd.DataFrame([[1, 2]], index=["z"], columns=cols)
        )
        assert_frame_equal(
            idx.bucket(df, (2, 4)),
            pd.DataFrame([[2, 4]], index=["x"], columns=cols)
        )
        assert_frame_equal(
            idx.bucket(df, (3, 4)),
            pd.DataFrame([[3, 4]], index=["c"], columns=cols)
        )

    def test_ignore_key_error(self):
        df = pd.DataFrame(
            [[1, 2], [3, 4]], columns=['a', 'b']
        )
        self.assertRaises(KeyError, lambda: ColumnsIndex('c').keys(df))
        self.assertEqual(ColumnsIndex(
            'c', ignore_key_error=True).keys(df), set())

    def test_index_elements(self):
        cols = ['col1', 'col2']
        df = pd.DataFrame(
            [
                [['a', 'b'], 'q'],
                [['c'], 'w'],
                [['b'], 'e'],
            ],
            index=RangeIndex(start=0, stop=3),
            columns=cols
        )
        idx = ColumnsIndex('col1', index_elements=True)
        keys = idx.keys(df)
        self.assertEqual(keys, set([('a',), ('b',), ('c',)]))
        assert_frame_equal(
            idx.bucket(df, ('a',)),
            pd.DataFrame([
                [['a', 'b'], 'q']
            ], index=[0], columns=cols)
        )
        assert_frame_equal(
            idx.bucket(df, ('b',)),
            pd.DataFrame([
                [['a', 'b'], 'q'],
                [['b'], 'e'],
            ], index=[0, 2], columns=cols)
        )

    def test_index_elements_multi_columns(self):
        cols = ['col1', 'col2', 'col3']
        df = pd.DataFrame(
            [
                [['a', 'b'], 'q', [1]],
                [['c'], 'w', [2, 3]],
                [['b'], 'e', [1]],
            ],
            index=RangeIndex(start=0, stop=3),
            columns=cols
        )
        idx = ColumnsIndex(['col1', 'col3'], index_elements=True)
        keys = idx.keys(df)
        self.assertEqual(keys, set([
            ('c', 2), ('a', 1), ('b', 1), ('b', 1), ('c', 3)
        ]))
        assert_frame_equal(
            idx.bucket(df, ('a', 1)),
            pd.DataFrame([
                [['a', 'b'], 'q', [1]],
            ], index=[0], columns=cols)
        )
        assert_frame_equal(
            idx.bucket(df, ('b', 1)),
            pd.DataFrame([
                [['a', 'b'], 'q', [1]],
                [['b'], 'e', [1]],
            ], index=[0, 2], columns=cols)
        )


class MultiIndexTestCase(BaseIndexTestCase):
    def test_index(self):
        cols = ["c", "d"]
        df = pd.DataFrame(
            [[1, 2], [2, 4], [3, 4]], index=["x", "y", "z"], columns=cols
        )

        idx = MultiIndex([
            ColumnsIndex('c'),
            ColumnsIndex('d')
        ])
        keys = idx.keys(df)
        self.assertEqual(keys, set([(1,), (2,), (3,), (4,)]))
        assert_frame_equal(
            idx.bucket(df, (1,)),
            pd.DataFrame([[1, 2]], index=["x"], columns=cols)
        )
        assert_frame_equal(
            idx.bucket(df, (2,)),
            pd.DataFrame([[1, 2], [2, 4]], index=["x", "y"], columns=cols)
        )
        assert_frame_equal(
            idx.bucket(df, (3,)),
            pd.DataFrame([[3, 4]], index=["z"], columns=cols)
        )
        assert_frame_equal(
            idx.bucket(df, (4,)),
            pd.DataFrame([[2, 4], [3, 4]], index=["y", "z"], columns=cols)
        )

        idx = MultiIndex([
            ColumnsIndex('c'),
            ColumnsIndex('d')
        ], combine_keys=True)
        keys = idx.keys(df)
        self.assertEqual(keys, set([
            ((3,), (4,)),
            ((2,), (4,)),
            ((1,), (2,)),
        ]))
        assert_frame_equal(
            idx.bucket(df, ((1,), (2,))),
            pd.DataFrame([[1, 2]], index=["x"], columns=cols)
        )
        assert_frame_equal(
            idx.bucket(df, ((2,), (4,))),
            pd.DataFrame([[2, 4]], index=["y"], columns=cols)
        )
        assert_frame_equal(
            idx.bucket(df, ((3,), (4,))),
            pd.DataFrame([[3, 4]], index=["z"], columns=cols)
        )



"""
An index divides the data up into one or more buckets. Only records in the same bucket are then matched against each other.
When used correctly, indexing decreases the number of pairs to compare and speeds up the matching process significantly.
"""

import operator
import functools
import itertools
from collections.abc import Iterable
from typing import Type
from abc import ABC, abstractmethod

import pandas as pd


class BaseIndex(ABC):
    """Abstract base class for all index classes.

    Sub-class should implement :meth:`_key_ind_map` method.

    .. automethod:: _key_ind_map
    """
    _dfs: list[tuple[pd.DataFrame, list]]

    def __init__(self) -> None:
        self._dfs = list()

    @abstractmethod
    def _key_ind_map(self, df: pd.DataFrame) -> dict:
        """Returns a mapping between bucket keys and row indices.

        :param df: the data to index.
        :type df: :class:`pandas:pandas.DataFrame`

        :return: A mapping between bucket key and all row indices that belong
            to the bucket. Key could be anything hashable but the value must always
            be a list even if there is only one row.
        :rtype: :obj:`dict`
        """
        raise NotImplementedError()

    def keys(self, df: pd.DataFrame) -> set:
        """Returns a set of keys that could be used to retrieve buckets.

        :param df: the data to index.
        :type df: :class:`pandas:pandas.DataFrame`

        :return: A set of bucket keys.
        :rtype: :obj:`set`
        """
        key_inds = self._key_ind_map(df)
        self._dfs.append((df, key_inds))
        return set(key_inds.keys())

    def bucket(self, df: pd.DataFrame, key: any) -> pd.DataFrame:
        """Retrieves a bucket given the original data and a bucket key.

        :param df: the data to index.
        :type df: :class:`pandas:pandas.DataFrame`

        :param key: one of the keys returned from :meth:`BaseIndex.keys`.
        :type key: any

        :return: Rows in bucket.
        :rtype: :class:`pandas:pandas.DataFrame`
        """
        for frame, key_inds in self._dfs:
            if frame is df:
                break
        else:
            raise ValueError(
                "frame not registered, make sure to run index.keys(df) first."
            )
        rows = df.loc[key_inds[key]]
        if len(rows.shape) == 1:
            rows = rows.to_frame().transpose()
        return rows


class NoopIndex(BaseIndex):
    """Returns all data as a single bucket.

    Using this is like using no index at all. Useful for when you don't care about
    performance (e.g. when there are not too many rows).
    """

    def _key_ind_map(self, df: pd.DataFrame) -> dict:
        return {0: df.index.tolist()}


class ColumnsIndex(BaseIndex):
    """Split data into multiple buckets based on one or more columns.
    """

    def __init__(self, cols: str or list[str], ignore_key_error: bool = False, index_elements: bool = False) -> None:
        """
        :param cols: single column name or list of column names to index.
        :type cols: :obj:`str` or :obj:`list` of :obj:`str`

        :param ignore_key_error: When set to True, a column does not exist in the frame, don't produce
            any bucket instead of raising a KeyError.
        :type ignore_key_error: :obj:`bool`

        :param index_elements: Set this to True when each value in the column to index is a list, and
            you want to index using the list elements.
        :type index_elements: :obj:`bool`
        """
        super().__init__()
        self._ignore_key_error = ignore_key_error
        self._index_elements = index_elements
        if type(cols) is str:
            self._cols = [cols]
        else:
            self._cols = cols

    def _key_ind_map(self, df: pd.DataFrame) -> dict:
        result = dict()
        try:
            for idx, row in df.iterrows():
                if self._index_elements:
                    for col in self._cols:
                        if not isinstance(row[col], Iterable):
                            raise ValueError('column %s at row %s is not iterable: %s' % (
                                col, row.name, row[col]
                            ))
                    for key in itertools.product(*list(row[col] for col in self._cols)):
                        result.setdefault(key, list()).append(idx)
                else:
                    key = tuple(
                        row[col] for col in self._cols
                    )
                    result.setdefault(key, list()).append(idx)
            for l in result.values():
                l.sort()
        except KeyError:
            if not self._ignore_key_error:
                raise
        return result


class MultiIndex(BaseIndex):
    """Creates bucket keys by combining bucket keys from two or more indices.

    This has two modes of operation:

    - When **combine_keys** is `False`: the key sets of each index are concatenated together, this is like OR-ing the keys.
    - When **combine_keys** is `True`: the final key set is the cartesian product of all key sets, this is like AND-ing the keys.
    """

    def __init__(self, indices: list[Type[BaseIndex]], combine_keys: bool = False) -> None:
        """
        :param indices: list of indices to combine.
        :type indices: :obj:`list` of :class:`BaseIndex` subclass

        :param combine_keys: whether the final key set should be the cartesian product of all key sets, defaults to `False`.
        :type combine_keys: :obj:`bool`
        """
        super().__init__()
        self._indices = indices
        self._combine_keys = combine_keys

    def _key_ind_map(self, df: pd.DataFrame) -> dict:
        maps = [idx._key_ind_map(df) for idx in self._indices]
        result = dict()
        if self._combine_keys:
            for prod in itertools.product(*[d.items() for d in maps]):
                rows = list(functools.reduce(
                    operator.and_, [set(vals) for _, vals in prod]
                ))
                if len(rows) == 0:
                    continue
                key = tuple(key for key, _ in prod)
                result[key] = rows
        else:
            for d in maps:
                for key, rows in d.items():
                    result[key] = result.get(key, set()) | set(rows)
            result = {key: sorted(list(rows)) for key, rows in result.items()}
        return result



from datamatch.filters import DissimilarFilter, NonOverlappingFilter
import unittest

import pandas as pd
from pandas.testing import assert_frame_equal

from datamatch.indices import ColumnsIndex, NoopIndex
from datamatch.matchers import ThresholdMatcher
from datamatch.scorers import AbsoluteScorer, MaxScorer, SimSumScorer
from datamatch.similarities import JaroWinklerSimilarity, StringSimilarity
from datamatch.variators import Swap


class TestThresholdMatcher(unittest.TestCase):
    def test_match(self):
        cols = ["a", "b"]
        dfa = pd.DataFrame(
            [
                ["ab", "cd"],
                ["rtx", "qw"]
            ],
            columns=cols
        )
        dfb = pd.DataFrame(
            [
                ["ab", "cd"],
                ["ae", "vb"],
                ["rt", "qw"]
            ],
            columns=cols
        )
        matcher = ThresholdMatcher(
            NoopIndex(), {"a": StringSimilarity()}, dfa, dfb
        )

        self.assertEqual(matcher._pairs, [(0.8, 1, 2), (1.0, 0, 0)])

        self.assertEqual(
            matcher.get_index_pairs_within_thresholds(), [(1, 2), (0, 0)])

        assert_frame_equal(
            matcher.get_sample_pairs(),
            pd.DataFrame.from_records([
                {"score_range": "1.00-0.95", "pair_idx": 0,
                    "sim_score": 1.0, "row_key": 0, "a": "ab", "b": "cd"},
                {"score_range": "1.00-0.95", "pair_idx": 0,
                    "sim_score": 1.0, "row_key": 0, "a": "ab", "b": "cd"},
                {"score_range": "0.85-0.80", "pair_idx": 0,
                    "sim_score": 0.8, "row_key": 1, "a": "rtx", "b": "qw"},
                {"score_range": "0.85-0.80", "pair_idx": 0,
                    "sim_score": 0.8, "row_key": 2, "a": "rt", "b": "qw"},
            ], index=["score_range", "pair_idx", "sim_score", "row_key"])
        )

        assert_frame_equal(
            matcher.get_sample_pairs(include_exact_matches=False),
            pd.DataFrame.from_records([
                {"score_range": "0.85-0.80", "pair_idx": 0,
                    "sim_score": 0.8, "row_key": 1, "a": "rtx", "b": "qw"},
                {"score_range": "0.85-0.80", "pair_idx": 0,
                    "sim_score": 0.8, "row_key": 2, "a": "rt", "b": "qw"},
            ], index=["score_range", "pair_idx", "sim_score", "row_key"])
        )

        assert_frame_equal(
            matcher.get_all_pairs(),
            pd.DataFrame.from_records([
                {"pair_idx": 0, "sim_score": 1.0,
                    "row_key": 0, "a": "ab", "b": "cd"},
                {"pair_idx": 0, "sim_score": 1.0,
                    "row_key": 0, "a": "ab", "b": "cd"},
                {"pair_idx": 1, "sim_score": 0.8,
                    "row_key": 1, "a": "rtx", "b": "qw"},
                {"pair_idx": 1, "sim_score": 0.8,
                    "row_key": 2, "a": "rt", "b": "qw"},
            ], index=["pair_idx", "sim_score", "row_key"])
        )

        assert_frame_equal(
            matcher.get_all_pairs(include_exact_matches=False),
            pd.DataFrame.from_records([
                {"pair_idx": 1, "sim_score": 0.8,
                    "row_key": 1, "a": "rtx", "b": "qw"},
                {"pair_idx": 1, "sim_score": 0.8,
                    "row_key": 2, "a": "rt", "b": "qw"},
            ], index=["pair_idx", "sim_score", "row_key"])
        )

    def test_ensure_unique_index(self):
        dfa = pd.DataFrame(
            [[1, 2], [3, 4]], index=["a", "a"]
        )
        dfb = pd.DataFrame(
            [[5, 6], [7, 8]], index=["a", "b"]
        )
        with self.assertRaisesRegex(
                ValueError,
                "Dataframe index contains duplicates. Both frames need to have index free of duplicates."):
            ThresholdMatcher(NoopIndex(), {"a": StringSimilarity()}, dfa, dfb)

    def test_ensure_same_columns(self):
        dfa = pd.DataFrame(
            [[1, 2], [3, 4]], columns=["a", "c"]
        )
        dfb = pd.DataFrame(
            [[5, 6], [7, 8]], columns=["a", "b"]
        )
        with self.assertRaisesRegex(
                ValueError,
                "Dataframe columns are not equal."):
            ThresholdMatcher(NoopIndex(), {"a": StringSimilarity()}, dfa, dfb)

    def test_deduplicate(self):
        cols = ['last', 'first']
        df = pd.DataFrame([
            ['beech', 'freddie'],
            ['beech', 'freedie'],
            ['dupas', 'demia'],
            ['dupas', 'demeia'],
            ['brown', 'latoya'],
            ['bowen', 'latoya'],
            ['rhea', 'cherri'],
            ['rhea', 'cherrie'],
            ['be', 'freedie'],
            ['du', 'demeia'],
            ['teneisha', 'green'],
            ['tyler', 'green'],
            ['te neisha', 'green'],
            ['t', 'green'],
        ], columns=cols)

        matcher = ThresholdMatcher(NoopIndex(), {
            'last': JaroWinklerSimilarity(),
            'first': JaroWinklerSimilarity()
        }, df)

        self.assertEqual(
            matcher.get_index_clusters_within_thresholds(0.83),
            [
                frozenset({6, 7}),
                frozenset({4, 5}),
                frozenset({2, 3, 9}),
                frozenset({10, 12, 13}),
                frozenset({0, 8, 1}),
            ],
        )

        self.maxDiff = None
        self.assertEqual(
            matcher.get_clusters_within_threshold(0.83).to_string(),
            '\n'.join([
                '                                             last    first',
                'cluster_idx pair_idx sim_score row_key                    ',
                '0           0        0.990522  6             rhea   cherri',
                '                               7             rhea  cherrie',
                '1           0        0.985297  10        teneisha    green',
                '                               12       te neisha    green',
                '            1        0.878609  10        teneisha    green',
                '                               13               t    green',
                '            2        0.876863  12       te neisha    green',
                '                               13               t    green',
                '2           0        0.980748  2            dupas    demia',
                '                               3            dupas   demeia',
                '            1        0.923472  3            dupas   demeia',
                '                               9               du   demeia',
                '            2        0.902589  2            dupas    demia',
                '                               9               du   demeia',
                '3           0        0.941913  4            brown   latoya',
                '                               5            bowen   latoya',
                '4           0        0.939581  0            beech  freddie',
                '                               1            beech  freedie',
                '            1        0.923472  1            beech  freedie',
                '                               8               be  freedie',
                '            2        0.857679  0            beech  freddie',
                '                               8               be  freedie',
            ]),
        )

    def test_swap_variator(self):
        cols = ['last', 'first']
        df = pd.DataFrame([
            ['blake', 'lauri'],
            ['lauri', 'blake'],
            ['robinson', 'alexis'],
            ['robertson', 'alexis'],
            ['haynes', 'terry'],
            ['terry', 'hayes']
        ], columns=cols)

        matcher = ThresholdMatcher(NoopIndex(), {
            'last': JaroWinklerSimilarity(),
            'first': JaroWinklerSimilarity()
        }, df, variator=Swap('first', 'last'))

        self.assertEqual(
            matcher.get_index_pairs_within_thresholds(),
            [(2, 3), (4, 5), (0, 1)]
        )

    def test_filters(self):
        cols = ['uid', 'first', 'agency', 'start', 'end']
        df = pd.DataFrame([
            ['1', 'john', 'slidell pd', 0, 10],
            ['2', 'john', 'slidell pd', 10, 20],
            ['3', 'john', 'slidell pd', 20, 30],
            ['4', 'john', 'gretna pd', 11, 21],
            ['5', 'john', 'gretna pd', 0, 7],
            ['6', 'john', 'gretna pd', 10, 18],
        ], columns=cols)

        matcher = ThresholdMatcher(NoopIndex(), {
            'first': JaroWinklerSimilarity()
        }, df, filters=[
            DissimilarFilter('agency'),
            NonOverlappingFilter('start', 'end')
        ])

        self.assertEqual(
            matcher.get_index_pairs_within_thresholds(),
            [(0, 3), (1, 4), (2, 4), (2, 5)]
        )

    def test_scorer(self):
        columns = ['first_name', 'attract_id']
        df = pd.DataFrame([
            ['john', 5],
            ['jim', 5],
            ['ted', 3],
            ['tedd', 2]
        ], columns=columns)

        matcher = ThresholdMatcher(NoopIndex(), MaxScorer([
            AbsoluteScorer('attract_id', 1),
            SimSumScorer({
                'first_name': JaroWinklerSimilarity()
            })
        ]), df)

        self.assertEqual(
            matcher.get_clusters_within_threshold().to_string(),
            '\n'.join([
                '                                       first_name  attract_id',
                'cluster_idx pair_idx sim_score row_key                       ',
                '0           0        1.000000  0             john           5',
                '                               1              jim           5',
                '1           0        0.941667  2              ted           3',
                '                               3             tedd           2',
            ]),
        )

        self.assertEqual(
            matcher.get_clusters_within_threshold(
                include_exact_matches=False).to_string(),
            '\n'.join([
                '                                       first_name  attract_id',
                'cluster_idx pair_idx sim_score row_key                       ',
                '1           0        0.941667  2              ted           3',
                '                               3             tedd           2',
            ]),
        )

    def test_func_scorer(self):
        self.maxDiff = None
        df = pd.DataFrame([
            ['j', 'john', 20],
            ['j', 'jim', 20],
            ['b', 'bill', 19],
            ['b', 'bob', 21]
        ], columns=['fc', 'name', 'age'])

        matcher = ThresholdMatcher(
            index=ColumnsIndex('fc'),
            scorer=lambda a, b: 1.0 if a.age == b.age else 0.8,
            dfa=df
        )

        self.assertEqual(
            matcher.get_clusters_within_threshold().to_string(),
            '\n'.join([
                '                                       fc  name  age',
                'cluster_idx pair_idx sim_score row_key              ',
                '0           0        1.0       0        j  john   20',
                '                               1        j   jim   20',
                '1           0        0.8       2        b  bill   19',
                '                               3        b   bob   21',
            ])
        )



"""
A filter discards pairs from the matching process. An index, which dictates which pair can be compared,
does the opposite. They are both employed to increase matching performance.
"""

from abc import ABC, abstractmethod

import pandas as pd


class BaseFilter(ABC):
    """Base class of all filter classes.

    Sub-class should implement the :meth:`valid` method.
    """

    @abstractmethod
    def valid(self, a: pd.Series, b: pd.Series) -> bool:
        """Returns true if a pair of records is valid (can be matched).

        :param a: the left record.
        :type a: :class:`pandas:pandas.Series`

        :param b: the right record.
        :type b: :class:`pandas:pandas.Series`

        :return: Whether these two records can be matched.
        :rtype: :obj:`bool`
        """
        raise NotImplementedError()


class DissimilarFilter(BaseFilter):
    """Eliminates pairs with the same value for a specific field.
    """

    def __init__(self, col: str, ignore_key_error: bool = False) -> None:
        """
        :param col: The column to check.
        :type col: :obj:`str`

        :param ignore_key_error: When set to True, if the column is not found, acts like a no-op
            filter instead of raising KeyError.
        :param ignore_key_error: :obj:`bool`
        """
        super().__init__()
        self._col = col
        self._ignore_key_error = ignore_key_error

    def valid(self, a: pd.Series, b: pd.Series) -> bool:
        """
        .. hiding method's docstring

        :meta private:
        """
        try:
            val_a = a[self._col]
            val_b = b[self._col]
        except KeyError:
            if self._ignore_key_error:
                return True
            raise
        if pd.isnull(val_a) or pd.isnull(val_b):
            return True
        return val_a != val_b


class NonOverlappingFilter(BaseFilter):
    """Eliminates pairs with overlapping ranges.

    This is usually used over time ranges, which ensures time exclusivity of a record.
    """

    def __init__(self, start: str, end: str) -> None:
        """
        Both start and end columns must be of the same type and must be comparable.

        e.g. `df[end] < df[start]` should produce a boolean series.

        :param start: the range start column.
        :type start: :obj:`str`

        :param end: the range end column.
        :type end: :obj:`str`
        """
        super().__init__()
        self._start = start
        self._end = end

    def valid(self, a: pd.Series, b: pd.Series) -> bool:
        """
        .. hiding method's docstring

        :meta private:
        """
        return a[self._end] < b[self._start] or a[self._start] > b[self._end]



from unittest import TestCase

import pandas as pd

from datamatch.filters import DissimilarFilter, NonOverlappingFilter


class DissimilarFilterTestCase(TestCase):
    def test_valid(self):
        f = DissimilarFilter('agency')
        index = ['agency', 'uid']
        self.assertFalse(f.valid(
            pd.Series(['slidell pd', '123'], index=index),
            pd.Series(['slidell pd', '456'], index=index)
        ))
        self.assertTrue(f.valid(
            pd.Series(['gretna pd', '123'], index=index),
            pd.Series(['slidell pd', '456'], index=index)
        ))

    def test_ignore_key_error(self):
        index = ['agency', 'uid']
        series_a = pd.Series(['slidell pd', '123'], index=index)
        series_b = pd.Series(['slidell pd', '456'], index=index)
        self.assertRaises(
            KeyError,
            lambda: DissimilarFilter('first').valid(series_a, series_b)
        )
        self.assertTrue(DissimilarFilter(
            'first', ignore_key_error=True
        ).valid(series_a, series_b))


class NonOverlappingFilterTestCase(TestCase):
    def test_valid(self):
        f = NonOverlappingFilter('start', 'end')
        index = ['uid', 'start', 'end']
        self.assertFalse(f.valid(
            pd.Series(['123', 0, 4], index=index),
            pd.Series(['456', 3, 6], index=index)
        ))
        self.assertFalse(f.valid(
            pd.Series(['123', 10, 14], index=index),
            pd.Series(['456', 3, 16], index=index)
        ))
        self.assertFalse(f.valid(
            pd.Series(['123', 0, 4], index=index),
            pd.Series(['456', 3, 3], index=index)
        ))
        self.assertFalse(f.valid(
            pd.Series(['123', 10, 14], index=index),
            pd.Series(['456', 3, 11], index=index)
        ))
        self.assertTrue(f.valid(
            pd.Series(['123', 10, 10], index=index),
            pd.Series(['456', 3, 6], index=index)
        ))
        self.assertTrue(f.valid(
            pd.Series(['123', 10, 12], index=index),
            pd.Series(['456', 13, 16], index=index)
        ))
        self.assertTrue(f.valid(
            pd.Series(['123', 0, 10], index=index),
            pd.Series(['456', 23, 26], index=index)
        ))



from unittest import TestCase

import pandas as pd
import numpy as np

from datamatch.similarities import AbsoluteNumericalSimilarity, JaroWinklerSimilarity, RelativeNumericalSimilarity
from datamatch.scorers import AlterScorer, FuncScorer, RefuseToScoreException, SimSumScorer, AbsoluteScorer, MinScorer, MaxScorer


class SimSumScorerTestCase(TestCase):
    def test_score(self):
        scorer = SimSumScorer({
            'first_name': JaroWinklerSimilarity(),
            'age': AbsoluteNumericalSimilarity(10)
        })
        columns = ['first_name', 'age']
        self.assertEqual(
            scorer.score(
                pd.Series(["john", 41], index=columns),
                pd.Series(["john", 41], index=columns),
            ),
            1,
        )
        self.assertEqual(
            scorer.score(
                pd.Series(["jim", 41], index=columns),
                pd.Series(["jimm", 43], index=columns),
            ),
            0.8737093656105305,
        )


class AbsoluteScorerTestCase(TestCase):
    def test_score(self):
        scorer = AbsoluteScorer('attract_id', 1)
        self.assertEqual(
            scorer.score(
                pd.Series([1234], index=['attract_id']),
                pd.Series([1234], index=['attract_id']),
            ),
            1
        )
        self.assertRaises(RefuseToScoreException, lambda: scorer.score(
            pd.Series([1234], index=['attract_id']),
            pd.Series([2345], index=['attract_id']),
        ))
        self.assertRaises(RefuseToScoreException, lambda: scorer.score(
            pd.Series([1234], index=['attract_id']),
            pd.Series([np.NaN], index=['attract_id']),
        ))
        self.assertRaises(RefuseToScoreException, lambda: scorer.score(
            pd.Series([None], index=['attract_id']),
            pd.Series([1234], index=['attract_id']),
        ))

    def test_ignore_key_error(self):
        series_a = pd.Series([1], index=['a'])
        series_b = pd.Series([2], index=['a'])
        self.assertRaises(
            KeyError,
            lambda: AbsoluteScorer('b', 1).score(series_a, series_b)
        )
        self.assertRaises(
            RefuseToScoreException,
            lambda: AbsoluteScorer(
                'b', 1, ignore_key_error=True
            ).score(series_a, series_b)
        )


class MaxScorerTestCase(TestCase):
    def test_score(self):
        scorer = MaxScorer([
            AbsoluteScorer('attract_id', 1),
            SimSumScorer({
                'first_name': JaroWinklerSimilarity()
            })
        ])
        columns = ['first_name', 'attract_id']
        self.assertEqual(
            scorer.score(
                pd.Series(['john', 5], index=columns),
                pd.Series(['jim', 5], index=columns),
            ),
            1
        )
        self.assertEqual(
            scorer.score(
                pd.Series(['john', 5], index=columns),
                pd.Series(['jim', 4], index=columns),
            ),
            0.575
        )


class MinScorerTestCase(TestCase):
    def test_score(self):
        scorer = MinScorer([
            AbsoluteScorer('repell_id', 0),
            SimSumScorer({
                'first_name': JaroWinklerSimilarity()
            })
        ])
        columns = ['first_name', 'repell_id']
        self.assertEqual(
            scorer.score(
                pd.Series(['john', 5], index=columns),
                pd.Series(['jim', 5], index=columns),
            ),
            0
        )
        self.assertEqual(
            scorer.score(
                pd.Series(['john', 5], index=columns),
                pd.Series(['jim', 4], index=columns),
            ),
            0.575
        )


class AlterScorerTestCase(TestCase):
    def test_score(self):
        scorer = AlterScorer(
            scorer=SimSumScorer({
                'name': JaroWinklerSimilarity(),
                'age': RelativeNumericalSimilarity(100)
            }),
            values=pd.Series(
                [1, 2, 1, 3],
                index=[1, 2, 3, 4]
            ),
            alter=lambda x: x/2
        )
        columns = ['name', 'age']
        self.assertEqual(
            scorer.score(
                pd.Series(['john', 20], index=columns, name=2),
                pd.Series(['jim', 21], index=columns, name=4)
            ),
            0.7866557310723826
        )
        self.assertEqual(
            scorer.score(
                pd.Series(['john', 20], index=columns, name=1),
                pd.Series(['jim', 21], index=columns, name=3)
            ),
            0.3933278655361913
        )
        self.assertEqual(
            scorer.score(
                pd.Series(['john', 20], index=columns, name=5),
                pd.Series(['jim', 21], index=columns, name=6)
            ),
            0.7866557310723826
        )


class FuncScorerTestCase(TestCase):
    def test_score(self):
        scorer = FuncScorer(lambda a, b: 1.0 if a.age == b.age else 0.5)
        columns = ['name', 'age']
        self.assertEqual(
            scorer.score(
                pd.Series(['john', 20], index=columns),
                pd.Series(['jim', 21], index=columns)
            ),
            0.5
        )
        self.assertEqual(
            scorer.score(
                pd.Series(['john', 20], index=columns),
                pd.Series(['jim', 20], index=columns)
            ),
            1.0
        )



"""
A scorer decide how the similarity score for each pair is calculated.
"""

from abc import ABC, abstractmethod
import inspect
import math
from typing import Callable, Type

import pandas as pd


class RefuseToScoreException(Exception):
    """Raise to delegate scoring to a parent object
    """
    pass


ScoreFunc = Callable[[pd.Series, pd.Series], float]


class BaseScorer(ABC):
    """Base class for all scorer classes.

    Sub-class should implement the :meth:`score` method.
    """

    @abstractmethod
    def score(self, a: pd.Series, b: pd.Series) -> float:
        """Returns similarity score (0 <= sim <= 1) for a pair of records.

        :param a: The left record.
        :type a: :class:`pandas:pandas.Series`

        :param b: The right record.
        :type b: :class:`pandas:pandas.Series`

        :return: Similarity score.
        :rtype: :obj:`float`
        """
        raise NotImplementedError()


class SimSumScorer(BaseScorer):
    """Returns the sum of similarity values of all fields.
    """

    def __init__(self, fields: dict) -> None:
        """
        :param fields: The mapping between field name and :ref:`similarity class <Similarities>` to use.
        :type fields: :obj:`dict` of similarity classes
        """
        super().__init__()
        self._fields = fields

    def score(self, a: pd.Series, b: pd.Series) -> float:
        """
        .. hiding method's docstring

        :meta private:
        """
        sim_vec = dict()
        for k, scls in self._fields.items():
            if pd.isnull(a[k]) or pd.isnull(b[k]):
                sim_vec[k] = 0
            else:
                sim_vec[k] = scls.sim(a[k], b[k])
        return math.sqrt(
            sum(v * v for v in sim_vec.values()) / len(self._fields))


class AbsoluteScorer(BaseScorer):
    """Returns an arbitrary score if both records have the same value for a column.

    If the values are not equal or one of them is null then this scorer will raise :class:`RefuseToScoreException`.
    Therefore, this class should never be used on its own but always wrapped in either
    :class:`MaxScorer` or :class:`MinScorer`.
    """

    def __init__(self, column_name: str, score: float, ignore_key_error: bool = False) -> None:
        """
        :param column_name: The column to compare.
        :type column_name: :obj:`str`

        :param score: The score to return.
        :type score: :obj:`float`

        :param ignore_key_error: When set to True, if the column does not exist in either record,
            raise RefuseToScoreException (delegate to a parent scorer) instead of KeyError.
        :type ignore_key_error: :obj:`bool`
        """
        self._column = column_name
        self._default_score = score
        self._ignore_key_error = ignore_key_error

    def score(self, a: pd.Series, b: pd.Series) -> float:
        """
        .. hiding method's docstring

        :meta private:
        """
        try:
            if pd.isnull(a[self._column]) or pd.isnull(b[self._column]):
                raise RefuseToScoreException("one of the values is null")
            elif a[self._column] == b[self._column]:
                return self._default_score
        except KeyError:
            if self._ignore_key_error:
                raise RefuseToScoreException(
                    "column does not exist in one of the record")
            else:
                raise
        raise RefuseToScoreException("values are not equal")


class MaxScorer(BaseScorer):
    """Returns the max value from the scores of all child scorers.
    """

    def __init__(self, scorers: list[Type[BaseScorer]]) -> None:
        """
        :param scorers: The children classes.
        :type scorers: list of :class:`BaseScorer` subclasses.
        """
        self._scorers = scorers

    def score(self, a: pd.Series, b: pd.Series) -> float:
        """
        .. hiding method's docstring

        :meta private:
        """
        val = None
        for scorer in self._scorers:
            try:
                n = scorer.score(a, b)
            except RefuseToScoreException:
                continue
            if val is None or n > val:
                val = n
        if val is None:
            raise RefuseToScoreException("all children refuse to score")
        return val


class MinScorer(BaseScorer):
    """Returns the min value from the scores of all child scorers.
    """

    def __init__(self, scorers: list[Type[BaseScorer]]) -> None:
        """
        :param scorers: The children classes.
        :type scorers: list of :class:`BaseScorer` subclasses.
        """
        self._scorers = scorers

    def score(self, a: pd.Series, b: pd.Series) -> float:
        """
        .. hiding method's docstring

        :meta private:
        """
        val = None
        for scorer in self._scorers:
            try:
                n = scorer.score(a, b)
            except RefuseToScoreException:
                continue
            if val is None or n < val:
                val = n
        if val is None:
            raise RefuseToScoreException("all children refuse to score")
        return val


class AlterScorer(BaseScorer):
    """Modifies the score for pairs with the same given values
    """

    def __init__(self, scorer: Type[BaseScorer], values: pd.Series, alter: Callable[[int], int]) -> None:
        """
        :param scorer: The wrapped scorer.
        :type scorer: :class:`BaseScorer` subclass.
        :param values:
            for each pair, if both rows have index in this series and their values are the same then
            call `alter` to modify the final score.
        :type values: :class:`pandas:pandas.Series`
        :param alter: callback to modify the final score.
        :type alter: Callable[[int], int]
        """
        self._scorer = scorer
        self._values = values
        self._alter = alter

    def score(self, a: pd.Series, b: pd.Series) -> float:
        """
        .. hiding method's docstring

        :meta private:
        """
        score = self._scorer.score(a, b)
        try:
            if self._values[a.name] == self._values[b.name]:
                return self._alter(score)
        except KeyError:
            pass
        return score


class FuncScorer(BaseScorer):
    """Scores pairs by calling the given callback
    """

    def __init__(self, cb: ScoreFunc) -> None:
        """
        :param cb: Callback to calculate score
        :type cb: Callable[[:class:`pandas:pandas.Series`, :class:`pandas:pandas.Series`], :obj:`float`]
        """
        self._cb = cb

    def score(self, a: pd.Series, b: pd.Series) -> float:
        """
        .. hiding method's docstring

        :meta private:
        """
        return self._cb(a, b)



"""
When given a pair of values, a similarity class produces a similarity score that ranges between 0 and 1.
A similarity score of 1 means the two values are completely identical while 0 means there are no similarities.

Note that these classes only compute similarity scores between scalar values or native Python objects such as
:class:`datetime.datetime`, not the entire row (which is handled by :class:`ThresholdMatcher`).
"""

from datetime import datetime
from Levenshtein import ratio, jaro_winkler
from unidecode import unidecode


class StringSimilarity(object):
    """Computes a similarity score between two strings using Levenshtein distance.
    """

    def sim(self, a: str, b: str):
        """Returns a similarity score.

        :param a: The left string.
        :type a: :obj:`str`

        :param b: The right string.
        :type b: :obj:`str`

        :return: The similarity score.
        :rtype: :obj:`float`
        """
        return ratio(unidecode(a), unidecode(b))


class JaroWinklerSimilarity(object):
    """Similar to :class:`StringSimilarity` but gives extra weight to common prefixes.

    This class is very good at matching people's names because mistaking the first
    letter in a person's name should be a rare event.
    """

    def __init__(self, prefix_weight=0.1):
        """
        :param prefix_weight: The extra weight given to common prefixes, defaults to 0.1.
        :type prefix_weight: :obj:`float`
        """
        self._prefix_weight = prefix_weight

    def sim(self, a: str, b: str):
        """Returns a similarity score.

        :param a: The left string.
        :type a: :obj:`str`

        :param b: The right string.
        :type b: :obj:`str`

        :return: The similarity score.
        :rtype: :obj:`float`
        """
        return jaro_winkler(unidecode(a), unidecode(b), self._prefix_weight)


class AbsoluteNumericalSimilarity(object):
    """Computes similarity score between two numbers, extrapolated from a maximum absolute difference.

    Maximum absolute difference **d_max** (greater than 0) is the maximum tolerated difference
    between two numbers regardless of their actual values. If the difference between the two values
    are less than **d_max** then the similarity score between two values `a` and `b` is
    ``1.0 - abs(a - b) / d_max``. Otherwise, the score is 0.

    Implementation follows strategy for numerical values given in the Data Matching book [1]_
    """

    def __init__(self, d_max: float) -> None:
        """
        :param d_max: The maximum absolute difference.
        :type d_max: :obj:`float`
        """
        self._d_max = d_max

    def sim(self, a: float or int, b: float or int) -> float:
        """Returns a similarity score.

        :param a: The left number.
        :type a: :obj:`float` or :obj:`int`

        :param b: The right number.
        :type b: :obj:`float` or :obj:`int`

        :return: The similarity score.
        :rtype: :obj:`float`
        """
        d = abs(a - b)
        if d < self._d_max:
            return 1 - d / self._d_max
        return 0


class RelativeNumericalSimilarity(object):
    """Computes similarity score between two numbers, extrapolated from a maximum percentage difference.

    This class serves a similar purpose to :class:`AbsoluteNumericalSimilarity` but is more dependent on
    the actual values being compared.

    Percentage difference `pc` between two values `a` and `b` is defined as
    ``abs(a - b) / max(abs(a), abs(b)) * 100``.

    Maximum percentage difference **pc_max** (0 < pc_max < 100) is the maximum tolerated percentage
    difference between the two numbers. If the percentage difference `pc` is less than **pc_max** then
    the similarity score is calculated with ``1.0 - pc / pc_max``. Otherwise, the score is 0.

    Implementation follows strategy for numerical values given in the Data Matching book [1]_
    """

    def __init__(self, pc_max: int) -> None:
        """
        :param pc_max: The maximum percentage difference.
        :type pc_max: :obj:`int`
        """
        self._pc_max = pc_max

    def sim(self, a: float or int, b: float or int) -> float:
        """Returns a similarity score

        :param a: The left number.
        :type a: :obj:`float` or :obj:`int`

        :param b: The right number.
        :type b: :obj:`float` or :obj:`int`

        :return: The similarity score.
        :rtype: :obj:`float`
        """
        d = abs(a - b)
        pc = d / max(abs(a), abs(b)) * 100
        if pc < self._pc_max:
            return 1 - pc / self._pc_max
        return 0


class DateSimilarity(object):
    """Computes similarity score between two dates, extrapolated from a maximum absolute difference in days.

    Maximum absolute difference in days **d_max** is the maximum tolerated difference in days between two dates.
    Similar to :class:`AbsoluteNumericalSimilarity` if both dates `a` and `b` are less than **d_max** days
    apart then the similarity score is ``1 - (a - b) / d_max``.

    If however ``(a - b) >= d_max`` then we employs two alternative strategies to hedge against typos:

    - If the year values are the same but the month and day values are swapped, then the similarity score is 0.5.

    - | The last resort is to write each date in `YYYYMMDD` format and computes the similarity score between
      | two strings.

    Implementation follows strategy for date/time given in the Data Matching book [2]_
    """

    def __init__(self, d_max=30):
        """
        :param d_max: Dates that are less than this number of days apart will have similarity score as
            ``1 - <difference in days> / d_max``. For dates that are further apart, this class employs
            alternative methods to compute the similarity score to hedge against typos. This defaults to 30.
        :type d_max: :obj:`int`
        """
        self._d_max = d_max

    def sim(self, a: datetime, b: datetime):
        """Returns a similarity score.

        :param a: The left date.
        :type a: :obj:`datetime.datetime`

        :param b: The right date.
        :type b: :obj:`datetime.datetime`

        :return: The similarity score.
        :rtype: :obj:`float`
        """
        d = a - b
        if b > a:
            d = b - a
        if d.days < self._d_max:
            return 1 - d.days / self._d_max
        if a.year == b.year and a.month == b.day and a.day == b.month:
            return 0.5
        if a.year == b.year and a.day == b.day:
            return ratio(a.strftime("%Y%m%d"), b.strftime("%Y%m%d"))
        return 0



"""
A matcher matches records from two datasets (deduplicate if it is only given one dataset). It is also the main entry for this
library. Through its constructor, you can configure which index to use, which and how each column should be compared, etc...
"""

import operator
import functools
import itertools
import math
from bisect import bisect_left, bisect
from typing import Any, Iterator, Type

import pandas as pd
import numpy as np
from tqdm import tqdm

from .scorers import BaseScorer, FuncScorer, ScoreFunc, SimSumScorer
from .indices import BaseIndex
from .pairers import DeduplicatePairer, MatchPairer
from .filters import BaseFilter
from .variators import Variator


class ContinueOuter(Exception):
    """
    :meta private:
    """
    pass


MODE_MATCH = 1
MODE_DEDUP = 2


class ThresholdMatcher(object):
    """Matchs records by computing similarity score for each pair and discard those that fall below a threshold.

    This matcher does not require any training data, therefore it is perfect for when there is not too much data or if
    training data is not available.
    """

    _mode: int

    def __init__(
        self,
        index: Type[BaseIndex],
        scorer: dict or Type[BaseScorer] or ScoreFunc,
        dfa: pd.DataFrame,
        dfb: pd.DataFrame or None = None,
        variator: Type[Variator] or None = None,
        filters: list[Type[BaseFilter]] = [],
        show_progress: bool = False
    ) -> None:
        """
        If it is given two datasets then it will try to match records
        between them. If given only one dataset then it attempts to detect
        duplicates instead.

        :param index: The index to divide the dataset into distinct buckets.
        :type index: sub-class of :class:`datamatch.indices.BaseIndex`

        :param scorer: The scorer class to score each pair. If it is a dict then create
            a :class:`datamatch.scorers.SimSumScorer` with that dict and use it.
        :type scorer: Callable[[:class:`pandas:pandas.Series`, :class:`pandas:pandas.Series`], :obj:`float`] or sub-class of :class:`datamatch.scorers.BaseScorer` or :obj:`dict` of similarity classes

        :param dfa: The left dataset to match. Its index must not contain duplicates.
        :type dfa: :class:`pandas:pandas.DataFrame`

        :param dfb: The right dataset to match. Its index must not contain duplicates and
            its column must match **dfa**'s. If this is not given then the matcher will
            attempt to deduplicate **dfa** instead.
        :type dfb: :class:`pandas:pandas.DataFrame`, optional

        :param variator: The :ref:`variator <variators>` to use.
        :type variator: sub-class of :class:`datamatch.variators.Variator`, optional

        :param filters: The list of :ref:`filters` to use.
        :type filters: :obj:`list` of sub-class of :class:`datamatch.filters.BaseFilter`, optional

        :param show_progress: Prints a `tqdm <https://github.com/tqdm/tqdm>`_ progress bar to console during matching, defaults to False.
        :type show_progress: :obj:`bool`, optional
        """
        if dfb is None:
            self._pairer = DeduplicatePairer(dfa, index)
            self._mode = MODE_DEDUP
        else:
            self._pairer = MatchPairer(dfa, dfb, index)
            self._mode = MODE_MATCH
        if type(scorer) is dict:
            self._scorer = SimSumScorer(scorer)
        elif type(scorer).__name__ == 'function':
            self._scorer = FuncScorer(scorer)
        else:
            self._scorer = scorer
        if variator is None:
            self._variator = Variator()
        else:
            self._variator = variator
        self._filters = filters
        self._show_progress = show_progress
        self._score_all_pairs()

    def _remove_lesser_matches(self):
        """If a row already have a better match then remove the other matches.
        """
        indices_a = set()
        indices_b = set()
        keep = []
        self._pairs.reverse()
        for i, tup in enumerate(self._pairs):
            _, idx_a, idx_b = tup
            if idx_a in indices_a or idx_b in indices_b:
                continue
            indices_a.add(idx_a)
            indices_b.add(idx_b)
            keep.append(i)
        self._pairs = [self._pairs[i] for i in keep]
        self._pairs.reverse()

    def _valid_pairs(self) -> Iterator:
        for (idx_a, rec_a), (idx_b, rec_b) in self._pairer.pairs():
            try:
                for f in self._filters:
                    if not f.valid(rec_a, rec_b):
                        raise ContinueOuter
            except ContinueOuter:
                continue
            yield (idx_a, rec_a), (idx_b, rec_b)

    def _score_all_pairs(self):
        """Calculate similarity value for all pairs of records.
        """
        pairs = []
        valid_pairs = self._valid_pairs()
        if self._show_progress:
            valid_pairs = tqdm(valid_pairs, desc='scoring pairs')
        for (idx_a, rec_a), (idx_b, rec_b) in valid_pairs:
            sim = max(
                self._scorer.score(ser_a, ser_b)
                for ser_a, ser_b in itertools.product(
                    self._variator.variations(rec_a),
                    self._variator.variations(rec_b)
                )
            )
            pairs.append((sim, idx_a, idx_b))
        self._pairs = sorted(pairs, key=operator.itemgetter(0))
        # in dedup mode we can group more than two records therefore we're not dropping lesser matches
        if self._mode == MODE_MATCH:
            self._remove_lesser_matches()
        self._scores = [t[0] for t in self._pairs]

    def _split_clusters(self, orig_cluster: set[tuple[float, Any, Any]]) -> list[tuple[frozenset, set]]:
        paths: dict[Any, set] = {}
        pairs: dict[frozenset, tuple[float, Any, Any]] = {}
        nodes = set()
        for sim, idx_a, idx_b in orig_cluster:
            paths.setdefault(idx_a, set()).add(idx_b)
            paths.setdefault(idx_b, set()).add(idx_a)
            nodes.add(idx_a)
            nodes.add(idx_b)
            pairs[frozenset([idx_a, idx_b])] = (sim, idx_a, idx_b)
        clusters: list[set] = []
        clustered = set()
        for node in nodes:
            if node in clustered:
                continue
            cluster = set([node])
            clustered.add(node)
            queue = [node]
            # BFS to find all members of cluster
            while len(queue) > 0:
                cur = queue.pop()
                for neighbor in paths[cur]:
                    if neighbor in clustered:
                        continue
                    if all([n in paths[neighbor] for n in cluster]):
                        clustered.add(neighbor)
                        cluster.add(neighbor)
                        queue.append(neighbor)
            clusters.append(cluster)
        return [
            (
                frozenset(s),
                set([
                    pairs[frozenset(y)]
                    for y in itertools.combinations(s, 2)
                ])
            )
            for s in clusters if len(s) > 1
        ]

    def _get_clusters_dict_within_thresholds(self, lower_bound=0.7, upper_bound=1) -> dict[frozenset, set]:
        pairs = self._pairs[
            bisect_left(self._scores, lower_bound):
            bisect(self._scores, upper_bound)]
        pairs.reverse()
        clusters: dict[frozenset, set] = {}

        for sim, idx_a, idx_b in pairs:
            cluster_keys: list[set] = []
            for key in clusters:
                if idx_a in key or idx_b in key:
                    cluster_keys.append(key)
            new_key = set()
            new_val = set()
            for key in cluster_keys:
                new_key = new_key.union(key)
                new_val = new_val.union(clusters[key])
                clusters.pop(key, None)

            new_key.add(idx_a)
            new_key.add(idx_b)
            new_val.add((sim, idx_a, idx_b))
            clusters.__setitem__(frozenset(new_key), new_val)

        return dict(functools.reduce(operator.add, [
            self._split_clusters(cluster) for cluster in clusters.values()
        ]))

    def get_index_clusters_within_thresholds(self, lower_bound=0.7, upper_bound=1) -> list[frozenset]:
        """Returns index clusters with similarity scores within the specified thresholds.

        :param lower_bound: The lower threshold below which pairs won't be included, defaults to 0.7.
        :type lower_bound: :obj:`float`

        :param upper_bound: The upper threshold above which pairs won't be included, defaults to 1.
        :type upper_bound: :obj:`float`

        :return: A list of clusters, each cluster is a set of indices.
        :rtype: :obj:`list` of :obj:`frozenset`
        """
        return [
            idx_set for idx_set in
            self._get_clusters_dict_within_thresholds(lower_bound, upper_bound)
        ]

    def get_clusters_within_threshold(self, lower_bound=0.7, upper_bound=1, include_exact_matches: bool = True) -> pd.DataFrame:
        """Returns all clusters between a lower bound and upper bound.

        :param lower_bound: The lower threshold below which pairs won't be included, defaults to 0.7.
        :type lower_bound: :obj:`float`

        :param upper_bound: The upper threshold above which pairs won't be included, defaults to 1.
        :type upper_bound: :obj:`float`

        :param include_exact_matches: Includes clusters with score = 1.0.
        :type include_exact_matches: :obj:`bool`

        :return: A multi-indexed frame that contains all matched clusters.
        :rtype: :class:`pandas:pandas.DataFrame`
        """
        records = []
        clusters = sorted([
            sorted(pairs_set, key=lambda x: x[0], reverse=True)
            for pairs_set in self._get_clusters_dict_within_thresholds(
                lower_bound, upper_bound
            ).values()
        ], key=lambda x: x[0][0], reverse=True)
        for cluster_idx, cluster in enumerate(clusters):
            if not include_exact_matches and sum(score for score, _, _ in cluster) == len(cluster):
                continue
            for pair_idx, pair in enumerate(cluster):
                sim_score, idx_a, idx_b = pair
                records.append(dict([
                    ("cluster_idx", cluster_idx), ("pair_idx", pair_idx),
                    ("sim_score", sim_score), ("row_key", idx_a)
                ] + list(self._pairer.frame_a.loc[idx_a].to_dict().items())))
                records.append(dict([
                    ("cluster_idx", cluster_idx), ("pair_idx", pair_idx),
                    ("sim_score", sim_score), ("row_key", idx_b)
                ] + list(self._pairer.frame_b.loc[idx_b].to_dict().items())))
        return pd.DataFrame.from_records(
            records,
            index=["cluster_idx", "pair_idx", "sim_score", "row_key"])

    def get_index_pairs_within_thresholds(self, lower_bound=0.7, upper_bound=1) -> list:
        """Returns index pairs with similarity scores within specified thresholds.

        :param lower_bound: The lower threshold below which pairs won't be included, defaults to 0.7.
        :type lower_bound: :obj:`float`

        :param upper_bound: The upper threshold above which pairs won't be included, defaults to 1.
        :type upper_bound: :obj:`float`

        :return: A list of pairs of indices of matching records.
        :rtype: :obj:`list` of :obj:`tuple`
        """
        return [t[1:] for t in self._pairs[
            bisect_left(self._scores, lower_bound):
            bisect(self._scores, upper_bound)]]

    def get_sample_pairs(self, sample_counts=5, lower_bound=0.7, upper_bound=1, step=0.05, include_exact_matches: bool = True) -> pd.DataFrame:
        """Returns samples of record pairs for each range of similarity scores.

        :param sample_counts: The number of samples in each range.
        :type sample_counts: :obj:`int`

        :param lower_bound: The lower threshold below which pairs won't be included, defaults to 0.7.
        :type lower_bound: :obj:`float`

        :param upper_bound: The upper threshold above which pairs won't be included, defaults to 1.
        :type upper_bound: :obj:`float`

        :param step: The width of each range.
        :type step: :obj:`float`

        :param include_exact_matches: Includes pairs with score = 1.0.
        :type include_exact_matches: :obj:`bool`

        :return: A multi-indexed frame that only contain samples for each range.
        :rtype: :class:`pandas:pandas.DataFrame`
        """
        sample_records = []
        ranges = list(np.arange(upper_bound, lower_bound, -step)
                      ) + [lower_bound]
        for i, upper_val in enumerate(ranges[:-1]):
            lower_val = ranges[i+1]
            score_range = "%.2f-%.2f" % (upper_val, lower_val)
            pairs = self._pairs[
                bisect(self._scores, lower_val):
                bisect(self._scores, upper_val)
            ][:sample_counts]
            pairs.reverse()
            for pair_idx, pair in enumerate(pairs):
                if not include_exact_matches and pair[0] == 1:
                    continue
                sim_score, idx_a, idx_b = pair
                sample_records.append(dict([
                    ("score_range", score_range), ("pair_idx", pair_idx),
                    ("sim_score", sim_score), ("row_key", idx_a)
                ] + list(self._pairer.frame_a.loc[idx_a].to_dict().items())))
                sample_records.append(dict([
                    ("score_range", score_range), ("pair_idx", pair_idx),
                    ("sim_score", sim_score), ("row_key", idx_b)
                ] + list(self._pairer.frame_b.loc[idx_b].to_dict().items())))
        return pd.DataFrame.from_records(
            sample_records,
            index=["score_range", "pair_idx", "sim_score", "row_key"])

    def get_all_pairs(self, lower_bound=0.7, upper_bound=1, include_exact_matches: bool = True) -> pd.DataFrame:
        """Returns all matching pairs between a lower bound and an upper bound.

        :param lower_bound: The lower threshold below which pairs won't be included, defaults to 0.7.
        :type lower_bound: :obj:`float`

        :param upper_bound: The upper threshold above which pairs won't be included, defaults to 1.
        :type upper_bound: :obj:`float`

        :param include_exact_matches: Includes pairs with score = 1.0.
        :type include_exact_matches: :obj:`bool`

        :return: A multi-indexed dataframe that only contain samples for each range.
        :rtype: :class:`pandas:pandas.DataFrame`
        """
        records = []
        pairs = reversed(self._pairs[
            bisect_left(self._scores, lower_bound):
            bisect(self._scores, upper_bound)])
        for pair_idx, pair in enumerate(pairs):
            sim_score, idx_a, idx_b = pair
            if not include_exact_matches and sim_score == 1:
                continue
            records.append(dict([
                ("pair_idx", pair_idx), ("sim_score", sim_score), ("row_key", idx_a)
            ] + list(self._pairer.frame_a.loc[idx_a].to_dict().items())))
            records.append(dict([
                ("pair_idx", pair_idx), ("sim_score", sim_score), ("row_key", idx_b)
            ] + list(self._pairer.frame_b.loc[idx_b].to_dict().items())))
        return pd.DataFrame.from_records(
            records,
            index=["pair_idx", "sim_score", "row_key"])

    def save_pairs_to_excel(
        self, name: str, match_threshold: float, sample_counts: int = 5, lower_bound: float = 0.7,
        step: float = 0.05, include_exact_matches: bool = True
    ) -> None:
        """Saves matching pairs to an Excel file.

        This will create an Excel file with 3 sheets:

        - Sample pairs: sample pairs for each score range, similar to output of :meth:`get_sample_pairs`.

        - All pairs: all pairs that score higher than **lower_bound** ordered by the similarity score.

        - Decision: selected threshold and how many pairs are counted as matched.

        :param name: The excel file to save to.
        :type name: :obj:`str`

        :param match_threshold: The score above which a pair is considered a match.
        :type match_threshold: :obj:`float`

        :param sample_counts: The number of samples per score range in the "Sample pairs" sheet.
        :type sample_counts: :obj:`int`

        :param lower_bound: The lower threshold below which pairs won't be included, defaults to 0.7.
        :type lower_bound: :obj:`float`

        :param step: The width of each range in the "Sample pairs" sheet.
        :type step: :obj:`float`

        :param include_exact_matches: Includes pairs with score = 1.0.
        :type include_exact_matches: :obj:`bool`

        :rtype: :obj:`None`
        """
        with pd.ExcelWriter(name) as writer:
            self.get_sample_pairs(
                sample_counts, lower_bound, 1, step, include_exact_matches
            ).to_excel(writer, sheet_name='Sample pairs')
            self.get_all_pairs(
                lower_bound, 1, include_exact_matches
            ).to_excel(writer, sheet_name='All pairs')
            self._decision_series(
                match_threshold
            ).to_excel(writer, sheet_name="Decision")

    def _decision_series(self, match_threshold: float) -> pd.Series:
        dec = {
            "match_threshold": match_threshold,
            "number_of_matched_pairs": len(self._scores) - bisect_left(self._scores, match_threshold)
        }
        dec_tups = list(itertools.zip_longest(*dec.items()))
        return pd.Series(list(dec_tups[1]), index=list(dec_tups[0]))

    def save_clusters_to_excel(self, name: str, match_threshold: float, lower_bound: float = 0.7, include_exact_matches: bool = True) -> None:
        """Save matched clusters to an Excel file.

        This will create an Excel file with two sheets:

        - All clusters: all clusters that score higher than lower bound ordered by score.

        - Decision: selected threshold and how many pairs are counted as matched.

        :param lower_bound: The lower threshold below which pairs won't be included, defaults to 0.7.
        :type lower_bound: :obj:`float`

        :param name: The excel file to save to.
        :type name: :obj:`str`

        :param match_threshold: The score above which a pair is considered a match.
        :type match_threshold: :obj:`float`

        :param lower_bound: The lower threshold below which pairs won't be included, defaults to 0.7.
        :type lower_bound: :obj:`float`

        :param include_exact_matches: Includes clusters with score = 1.0.
        :type include_exact_matches: :obj:`bool`

        :rtype: :obj:`None`
        """
        with pd.ExcelWriter(name) as writer:
            self.get_clusters_within_threshold(
                lower_bound,
                include_exact_matches
            ).to_excel(writer, sheet_name='All clusters')
            self._decision_series(
                match_threshold
            ).to_excel(writer, sheet_name="Decision")

    def print_decision(self, match_threshold: float) -> None:
        """Prints number and percentage of matched pairs for selected threshold.

        :param match_threshold: The score above which a pair is considered a match.
        :type match_threshold: :obj:`float`

        :rtype: :obj:`None`
        """
        pairs = self.get_index_pairs_within_thresholds(
            lower_bound=match_threshold)
        num_pairs = len(pairs)
        print("for threshold %.3f:" % match_threshold)
        print("  %d matched pairs (%d%% of A, %d%% of B)" %
              (num_pairs, num_pairs / self._pairer.frame_a.shape[0] * 100, num_pairs / self._pairer.frame_b.shape[0] * 100))



from unittest import TestCase
import itertools

import pandas as pd
from pandas.testing import assert_series_equal

from datamatch.variators import Variator, Swap


class BaseVariatorTestCase(TestCase):
    def assert_series_list_equal(self, list_a, list_b):
        self.assertEqual(len(list_a), len(list_b))
        for rec_a, rec_b in itertools.zip_longest(list_a, list_b):
            assert_series_equal(rec_a, rec_b)


class VariatorTestCase(BaseVariatorTestCase):
    def test_variations(self):
        sr = pd.Series([1, 2, 3])
        v = Variator()
        self.assert_series_list_equal(list(v.variations(sr)), [sr])


class SwapTestCase(BaseVariatorTestCase):
    def test_variations(self):
        sr1 = pd.Series([1, 2, 3], index=['x', 'y', 'z'])
        sr2 = pd.Series([2, 2, 3], index=['x', 'y', 'z'])

        v = Swap('x', 'y')

        self.assert_series_list_equal(list(v.variations(sr1)), [
            sr1, pd.Series([2, 1, 3], index=['x', 'y', 'z'])
        ])

        self.assert_series_list_equal(list(v.variations(sr2)), [sr2])



"""
A pairer produces pairs of records for matching. Most users should never need
to touch a pairer but this is exposed anyway for the sake of customization.

There are two pairers corresponding to two strategies:

- | :class:`MatchPairer`: Takes in two datasets and produces pairs of records
  | such that each pair contain one record from one dataset and one record from
  | the other dataset. This pairer is utilized when :class:`ThresholdMatcher`
  | is given two datasets. It is useful for matching records between
  | two datasets.

- | :class:`DeduplicatePairer`: Takes in one dataset and produces pairs of records
  | each having only records from the input dataset. This pairer is utilized
  | when :class:`ThresholdMatcher` is given only one dataset. It is
  | useful for deduplication tasks.
"""

import itertools
from typing import Iterator, Type
from abc import ABC, abstractmethod

import pandas as pd

from .indices import BaseIndex


class BasePairer(ABC):
    """Abstract base class for all pairer classes.

    Sub-class must implement :meth:`frame_a`, :meth:`frame_b`, and :meth:`pairs`.

    :meth:`frame_a` should produce the left set of records, :meth:`frame_b`
    should produce the right set of records, whereas :meth:`pairs` should produce
    pairs of records (one from :meth:`frame_a`, one from :meth:`frame_b`).
    """

    @property
    @abstractmethod
    def frame_a(self) -> pd.DataFrame:
        """Returns the left set of records.

        :rtype: :class:`pandas:pandas.DataFrame`
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def frame_b(self) -> pd.DataFrame:
        """Returns the right set of records.

        :rtype: :class:`pandas:pandas.DataFrame`
        """
        raise NotImplementedError()

    @abstractmethod
    def pairs(self) -> Iterator:
        """Returns an iterator over pairs of records that should be compared.

        Each pair is a tuple of left record and right record. Each record is a tuple
        of two elements: the row index and the row data which is a :class:`pandas:pandas.Series`.

        :rtype: :ref:`Iterator <python:typeiter>`
        """
        raise NotImplementedError()


class MatchPairer(BasePairer):
    """Pairs records from two frames using the provided index.
    """

    def __init__(self, dfa: pd.DataFrame, dfb: pd.DataFrame, index: Type[BaseIndex]) -> None:
        """
        :param dfa: The left dataset.
        :type dfa: :class:`pandas:pandas.DataFrame`

        :param dfb: The right dataset.
        :type dfb: :class:`pandas:pandas.DataFrame`

        :param index: The index to divide datasets into buckets.
        :type index: sub-class of :class:`datamatch.indices.BaseIndex`
        """
        self._index = index
        if dfa.index.duplicated().any() or dfb.index.duplicated().any():
            raise ValueError(
                "Dataframe index contains duplicates. Both frames need to have index free of duplicates."
            )
        if set(dfa.columns) != set(dfb.columns):
            raise ValueError(
                "Dataframe columns are not equal."
            )

        self._dfa = dfa
        self._dfb = dfb

    @property
    def frame_a(self) -> pd.DataFrame:
        """
        .. hiding method's docstring

        :meta private:
        """
        return self._dfa

    @property
    def frame_b(self) -> pd.DataFrame:
        """
        .. hiding method's docstring

        :meta private:
        """
        return self._dfb

    def pairs(self) -> Iterator:
        """
        .. hiding method's docstring

        :meta private:
        """
        keys_a = self._index.keys(self._dfa)
        keys_b = self._index.keys(self._dfb)
        keys = list(keys_a.intersection(keys_b))
        for key in keys:
            rows_a = self._index.bucket(self._dfa, key)
            rows_b = self._index.bucket(self._dfb, key)
            for idx_a, ser_a in rows_a.iterrows():
                for idx_b, ser_b in rows_b.iterrows():
                    yield (idx_a, ser_a), (idx_b, ser_b)


class DeduplicatePairer(BasePairer):
    """Pairs records from a single frame for deduplication.

    As this class is only initialized with a single frame, both
    `frame_a` and `frame_b` returns this same frame.
    """

    def __init__(self, df: pd.DataFrame, index: Type[BaseIndex]) -> None:
        """
        :param df: The dataset to deduplicate.
        :type df: :class:`pandas:pandas.DataFrame`

        :param index: The index to divide datasets into buckets.
        :type index: sub-class of :class:`datamatch.indices.BaseIndex`
        """
        self._index = index
        if df.index.duplicated().any():
            raise ValueError(
                "Dataframe index contains duplicates."
            )
        self._df = df

    @property
    def frame_a(self) -> pd.DataFrame:
        """
        .. hiding method's docstring

        :meta private:
        """
        return self._df

    @property
    def frame_b(self) -> pd.DataFrame:
        """
        .. hiding method's docstring

        :meta private:
        """
        return self._df

    def pairs(self) -> Iterator:
        """
        .. hiding method's docstring

        :meta private:
        """
        keys = self._index.keys(self._df)
        for key in keys:
            rows = self._index.bucket(self._df, key)
            for row_a, row_b in itertools.combinations(rows.iterrows(), 2):
                yield row_a, row_b



from .indices import NoopIndex, ColumnsIndex, MultiIndex
from .similarities import (
    StringSimilarity, DateSimilarity, JaroWinklerSimilarity, AbsoluteNumericalSimilarity, RelativeNumericalSimilarity
)
from .matchers import ThresholdMatcher
from .variators import Variator, Swap
from .filters import DissimilarFilter, NonOverlappingFilter
from .scorers import (
    SimSumScorer, AbsoluteScorer, MinScorer, MaxScorer, RefuseToScoreException, AlterScorer
)

__all__ = [
    "NoopIndex", "ColumnsIndex", "JaroWinklerSimilarity", "StringSimilarity",
    "DateSimilarity", "ThresholdMatcher", "MultiIndex", "Variator", "Swap",
    "DissimilarFilter", "NonOverlappingFilter", "AbsoluteNumericalSimilarity",
    "RelativeNumericalSimilarity",
    "SimSumScorer", "AbsoluteScorer", "MinScorer", "MaxScorer", "RefuseToScoreException", "AlterScorer"
]



"""
A variator creates variations of a single record. :class:`ThresholdMatcher` can produce different similarity
scores for different variations of the same record, discarding all but the highest score in the final result.
This is very useful in situations where values are not put into the correct columns (e.g. when a person's first
name and last name are swapped).
"""

from typing import Iterator

import pandas as pd


class Variator(object):
    """Base class of all variator classes.

    Sub-class should override method :meth:`variations`.

    This class also serves as a no-op variator. It simply returns the record as is.
    """

    def variations(self, sr: pd.Series) -> Iterator[pd.Series]:
        """Returns variations of the input record.

        :param sr: The input record.
        :type sr: :class:`pandas:pandas.Series`

        :rtype: :ref:`Iterator <python:typeiter>` of :class:`pandas:pandas.Series`
        """
        yield sr


class Swap(Variator):
    """Produces variations by swapping values between two columns"""

    def __init__(self, column_a: str, column_b: str) -> None:
        """
        :param column_a: The left column.
        :type column_a: :obj:`str`

        :param column_b: The right column.
        :type column_b: :obj:`str`
        """
        super().__init__()
        self._col_a = column_a
        self._col_b = column_b

    def variations(self, sr: pd.Series) -> Iterator[pd.Series]:
        """
        .. hiding method's docstring

        :meta private:
        """
        yield sr
        if not (pd.isna(sr[self._col_a]) and pd.isna(sr[self._col_b])) and (sr[self._col_a] != sr[self._col_b]):
            sr = sr.copy(True)
            v = sr[self._col_a]
            sr.loc[self._col_a] = sr[self._col_b]
            sr.loc[self._col_b] = v
            yield sr



import unittest
from datetime import date

from datamatch.similarities import (
    StringSimilarity, DateSimilarity, JaroWinklerSimilarity, AbsoluteNumericalSimilarity, RelativeNumericalSimilarity
)


class TestStringSimilarity(unittest.TestCase):
    def test_sim(self):
        obj = StringSimilarity()
        self.assertEqual(obj.sim("abc", "abc"), 1)
        self.assertEqual(obj.sim("abc", "123"), 0)
        self.assertEqual(obj.sim("abce", "abcd"), 0.75)
        self.assertEqual(obj.sim("thang", "thăng"), 1)


class TestJaroWinklerSimilarity(unittest.TestCase):
    def test_sim(self):
        obj = JaroWinklerSimilarity(0.2)
        self.assertEqual(obj.sim("abc", "abc"), 1)
        self.assertEqual(obj.sim("abc", "123"), 0)
        self.assertEqual(obj.sim("abce", "abcd"), 0.9333333333333333)
        self.assertEqual(obj.sim("wbcd", "abcd"), 0.8333333333333334)


class TestDateSimilarity(unittest.TestCase):
    def test_sim(self):
        obj = DateSimilarity()
        self.assertEqual(obj.sim(date(2000, 10, 11), date(2000, 10, 11)), 1)
        # within 30 days difference
        self.assertEqual(obj.sim(date(2000, 10, 11), date(2000, 10, 5)), 0.8)
        self.assertEqual(
            obj.sim(date(2000, 10, 11), date(2000, 11, 5)), 0.16666666666666663)
        # completely different days
        self.assertEqual(obj.sim(date(2000, 10, 11), date(2001, 3, 15)), 0)
        # day & month is swapped
        self.assertEqual(obj.sim(date(2000, 9, 11), date(2000, 11, 9)), 0.5)
        # same year and day but month is different
        self.assertEqual(obj.sim(date(2000, 3, 20), date(2000, 8, 20)), 0.875)


class TestAbsoluteNumericalSimilarity(unittest.TestCase):
    def test_sim(self):
        obj = AbsoluteNumericalSimilarity(10)
        self.assertEqual(obj.sim(10, 10), 1)
        self.assertEqual(obj.sim(8.9, 8.9), 1)
        self.assertEqual(obj.sim(10, 5), 0.5)
        self.assertEqual(obj.sim(10, 15), 0.5)
        self.assertEqual(obj.sim(8.2, 3.1), 0.49)
        self.assertEqual(obj.sim(40, 10), 0)


class TestRelativeNumericalSimilarity(unittest.TestCase):
    def test_sim(self):
        obj = RelativeNumericalSimilarity(30)
        self.assertEqual(obj.sim(10000, 10000), 1)
        self.assertEqual(obj.sim(8.9, 8.9), 1)
        self.assertEqual(obj.sim(10000, 8500), 0.5)
        self.assertEqual(obj.sim(8500, 10000), 0.5)
        self.assertEqual(obj.sim(8.2, 3.1), 0)
        self.assertEqual(obj.sim(10000, 7000), 0)
